import logging
import asyncio
from typing import (
    Callable,
    List,
)
from pipeline.producers import get_producer


LOG = logging.getLogger(__name__)


def _on_connect(client, *args, **kwargs):
    LOG.debug('%s: CONNECTED', client._client_id)


async def _on_message(client, msg, *args, **kwargs):
    LOG.debug('RECV: %s', msg)


def _on_disconnect(client, *args, **kwargs):
    LOG.debug('%s: DISCONNECTED', client._client_id)


def _on_subscribe(client, *args, **kwargs):
    LOG.debug('%s: SUBSCRIBED TO TOPICS.', client._client_id)


class Client:
    ''' Base class for consumers.'''
    def __init__(
            self,
            uri: str,
            username: str = None,
            password: str = None,
            connected: bool = False,
            on_connect: Callable = _on_connect,
            on_message: Callable = _on_message,
            on_disconnect: Callable = _on_disconnect,
            on_subscribe: Callable = _on_subscribe,
    ):
        self.uri = uri
        self.username = username
        self.password = password
        self.connected = connected

        self.on_connect = on_connect
        self.on_message = on_message
        self.on_disconnect = on_disconnect
        self.on_subscribe = on_subscribe

        self.loop = asyncio.get_event_loop()

    async def _setup(self):
        self.producer = await get_producer()

    async def pipe_message(self, data: dict, target: str) -> None:
        ''' Pipes mesage to configured producer. '''
        await self.producer.produce_data(
            data=data,
            target=target,
        )

    async def connect(self, topics: List) -> None:
        raise NotImplementedError("No driver specified")

    async def disconnect(self) -> None:
        raise NotImplementedError("No driver specified")



from typing import Any

from .base import Client


async def init_client(cls: Client, **kwargs: Any) -> Client:
    '''
    Creates an instance and runs setup instructions of supplied client class.
    '''
    client = cls(**kwargs)
    await client._setup()
    return client





from pipeline import settings
from .base import Producer
from .kafka import KafkaProducer
from .mock import MockProducer
from .timescale import TimescaleProducer



async def create_producer(cls: Producer) -> Producer:
    '''
    Creates an instance and runs setup instructions of supplied producer class.
    '''
    producer = cls()
    await producer._setup()
    return producer


async def get_producer() -> Producer:
    ''' Returns chosen producer based applied config '''
    if settings.PRODUCER_DRIVER == 'kafka':
        return await create_producer(KafkaProducer)

    if settings.PRODUCER_DRIVER == 'timescale':
        return await create_producer(TimescaleProducer)

    return await create_producer(MockProducer)



# system config
LOG_DIR = None
LOG_LEVEL = 'DEBUG'

# Client settings
CLIENT_MQTT_TOPICS = None
CLIENT_KAFKA_TOPICS = None
CLIENT_URI = None
CLIENT_USERNAME = None
CLIENT_PASSWORD = None

# LoRa server settings
LORA_DRIVER = None

# Producer settings
PRODUCER_DRIVER = None
PRODUCER_TARGET = None
PRODUCER_HOSTS = None
PRODUCER_DATABASE = None
PRODUCER_USERNAME = None
PRODUCER_PASSWORD = None



import logging

from asyncpg import create_pool

from .base import Producer


LOG = logging.getLogger(__name__)


class TimescaleProducer(Producer):
    '''
    Uses connections from a shared connection pool to connect to a timescale database.

    Exposes a number of methods for ease of use implementation by a client.
    TODO: Document interface.
    '''
    async def _setup(self) -> None:
        await super()._setup()
        self.pool = await create_pool(
            host=self.hosts,
            database=self.database,
            user=self.username,
            password=self.password,
            loop=self.loop,
            min_size=10,
            max_size=100,
        )

    async def connect(self) -> None:
        self.connected = True

    async def disconnect(self) -> None:
        await self.producer.stop()
        self.connected = False

    async def produce_data(self, data: dict, target: str = None) -> None:
        '''
        Produces the values in the passed data dict in corresponding given order.
        '''
        async with self.pool.acquire() as connection:
            async with connection.transaction():
                await connection.execute(data)



import logging
import asyncio
import signal
from typing import Coroutine

from pipeline import STOP


LOG = logging.getLogger(__name__)


async def task_wrapper(task: Coroutine):
    ''' Wraps a coroutine as a task and passes the global stop signal to it '''
    try:
        await task(stop_signal=STOP)
    except asyncio.CancelledError:
        LOG.debug('Cancelled: %s', task.__name__)
    except asyncio.TimeoutError:
        LOG.debug('Timed out: %s. Restarting.', task.__name__)


def main(task):
    '''
    Main entry point for a task.
    Adds signal handlers and starts the task.
    '''
    loop = asyncio.get_event_loop()

    # connect signal handler for stopping program
    stop_signals = [signal.SIGINT, signal.SIGHUP]
    for value in stop_signals:
        loop.add_signal_handler(value, STOP.set)

    LOG.debug('Starting loop.')
    loop.run_until_complete(asyncio.gather(
        task_wrapper(task),
    ))

    LOG.debug('Main finished.')



import logging
import json

from aiokafka import AIOKafkaProducer

from .base import Producer


LOG = logging.getLogger(__name__)


class KafkaProducer(Producer):
    '''
    Connects to a kafka cluster and roduces messages as an encoded JSON structure.
    '''
    async def _setup(self) -> None:
        await super()._setup()
        self.producer = AIOKafkaProducer(
            loop=self.loop,
            bootstrap_servers=self.hosts
        )

    async def connect(self) -> None:
        LOG.debug('Connecting to Kafka stream.')
        await self.producer.start()
        self.connected = True

    async def disconnect(self) -> None:
        await self.producer.stop()
        self.connected = False

    async def produce_data(self, data: dict, target: str = None) -> None:
        LOG.debug('Trying to produce data: %s', data)

        if not self.connected:
            LOG.debug('Producer not connected')
            return

        await super().produce_data(data=data, target=target)
        await self.producer.send_and_wait(target, json.dumps(data).encode())



import logging
import json
from typing import (
    Tuple,
    Any,
)

from gmqtt import Client as GMQTTClient, Subscription
from gmqtt.mqtt.constants import MQTTv311

from .base import Client


LOG = logging.getLogger(__name__)


class MQTTClient(Client):
    ''' Client for subscribing to a broker and pipe incomming messages to configured producer. '''
    def __init__(
            self,
            **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self._inner_client = GMQTTClient(None)

        self._inner_client.on_connect = self.on_connect
        self._inner_client.on_message = self.on_message
        self._inner_client.on_disconnect = self.on_disconnect
        self._inner_client.on_subscribe = self.on_subscribe
        self._inner_client.pipe_message = self.pipe_message

        if not self.username is None:
            self._inner_client.set_auth_credentials(self.username, self.password)


    async def connect(self, topics: Tuple[str, int]) -> None:
        ''' Connects to broker. '''
        LOG.debug('Connecting to MQTT broker.')
        try:
            await self._inner_client.connect(self.uri, 1883, keepalive=60, version=MQTTv311)
            subscriptions = [Subscription(t[0], qos=t[1]) for t in topics]
            self._inner_client.subscribe(subscriptions, subscription_identifier=1)

            await self.producer.connect()

            self.connected = True
        except:
            self.connected = False


async def serialize(data):
    ''' Decodes recieved data to utf-8 and returns it as a dict. '''
    data = data.decode('utf-8')
    try:
        return json.loads(data)
    except json.decoder.JSONDecodeError:
        return data



import logging
from typing import (
    List,
    Any,
)
from aiokafka import AIOKafkaConsumer

from .base import Client


LOG = logging.getLogger(__name__)


def _on_subscribe(client, topics, **kwargs):
    LOG.debug('%s: SUBSCRIBED TO TOPICS: %s', client._client_id, topics)


class KafkaClient(Client):
    ''' Client for consuming records from a Kafka topic and
    pipe it to configured producer. '''
    def __init__(
            self,
            group_id: str,
            **kwargs: Any,
    ):
        super().__init__(on_subscribe=_on_subscribe, **kwargs)

        self.consumer = None
        self.group_id = group_id
        self.hosts = self.uri.split(',')

    async def connect(self, topics: List) -> None:
        self.consumer = AIOKafkaConsumer(
            *topics,
            group_id=self.group_id,
            bootstrap_servers=self.hosts,
            loop=self.loop,
        )

        await self.consumer.start()

        self.on_connect(self.consumer._client)
        self.on_subscribe(self.consumer._client, topics)
        self.connected = True

        try:
            async for msg in self.consumer:
                await self.on_message(self, msg) # TODO: A bit wierd syntax, fix.
        finally:
            self.disconnect()

    async def disconnect(self) -> None:
        self.consumer.stop()
        self.connected = False
        self.on_disconnect(self.consumer._client)



import logging
import asyncio

from pipeline import settings


LOG = logging.getLogger(__name__)


class Producer():
    ''' Base class for producers. '''
    def __init__(self):
        self.hosts = settings.PRODUCER_HOSTS
        self.target = settings.PRODUCER_TARGET
        self.database = settings.PRODUCER_DATABASE
        self.username = settings.PRODUCER_USERNAME
        self.password = settings.PRODUCER_PASSWORD

        self.producer = None
        self.connected = False
        self.loop = asyncio.get_event_loop()

    async def _setup(self) -> None:
        LOG.debug('Setting up producer.')

    async def connect(self) -> None:
        ''' Connect to producer destination '''
        raise NotImplementedError("No driver specified")

    async def disconnect(self) -> None:
        ''' Disconnect from producer destination '''
        raise NotImplementedError("No driver specified")

    async def produce_data(self, data: dict, target: str = None) -> None:
        '''
        Override this method to produce data to specific producer destination,
        which we randomly call target.
        '''
        if not target:
            target = settings.PRODUCER_TARGET



import logging

from .base import Producer


LOG = logging.getLogger(__name__)


class MockProducer(Producer):
    ''' Produces log messages instead of pushing an existing source '''
    async def connect(self) -> None:
        LOG.debug('Connected to mock producer')
        self.connected = True

    async def disconnect(self) -> None:
        self.connected = False

    async def produce_data(self, data: dict, target: str = None) -> None:
        LOG.debug('Mocking a message on topic %s: %s', data, target)



from datetime import datetime
import asyncio
import pytest

from pipeline.producers.timescale import _map_value, _map_values


@pytest.mark.asyncio
async def test_map_value():
    epoch_to_datetime = await _map_value(value=1581412003, value_type='timestamptz')
    assert datetime(2020, 2, 11, 10, 6, 43) == epoch_to_datetime

    str_to_text = await _map_value(value="abc123", value_type='text')
    assert "'abc123'" == str_to_text


@pytest.mark.asyncio
async def test_map_values():
    values = [
        (25.1, 'double'),
        ('abc123', 'text'),
        (1581412003, 'timestamptz')
    ]
    res = await _map_values(values=values)
    assert datetime(2020, 2, 11, 10, 6, 43) == epoch_to_datetime
